"""
client package contains the main class to interact with the SAT service.
"""

import json
import logging
from typing import Any, Callable, Dict, Optional, Union

import requests
from requests.exceptions import HTTPError

from py_sat.constant import (ACCESS_TOKEN_URL, ACCOUNT_PATH, CHECK_STATUS_PATH,
                             CHECKOUT_PATH, INQUIRY_PATH, PING_PATH,
                             PLAYGROUND_SAT_BASE_URL, PRODUCT_LIST_PATH)
from py_sat.exceptions import (GeneralException, InvalidInputException,
                               ResponseGeneralException,
                               UnauthenticatedException)
from py_sat.http_client import HTTPClient
from py_sat.models import (Account, ErrorResponse, InquiryRequest,
                           InquiryResponse, OrderDetail, OrderRequest,
                           PartnerProduct, PingResponse, ProductListResponse)
from py_sat.signature import Signature, SignatureType
from py_sat.utils import (generate_json_api_request,
                          parse_json_api_list_response,
                          parse_json_api_response)


class SATClientConfig:
    """
    SATClientConfig is a configuration class to initialize the SATClient
    """

    # Required
    client_id: str
    client_secret: str
    private_key: str

    # Optional
    public_key: Optional[str]
    padding_type: SignatureType
    is_debug: bool
    sat_base_url: str
    access_token_base_url: str
    timeout: int
    logger: logging.Logger

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        private_key: str,
    ):
        if not client_id or not isinstance(client_id, str):
            raise InvalidInputException("Client ID are required and must be a string")

        if not client_secret or not isinstance(client_secret, str):
            raise InvalidInputException(
                "Client Secret are required and must be a string"
            )

        if not private_key or not isinstance(private_key, str):
            raise InvalidInputException("Private key is required and must be a string")

        self.client_id = client_id
        self.client_secret = client_secret
        self.private_key = private_key

        self._set_default_value()

    def _set_default_value(self):
        logger: logging.Logger
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        self.logger = logger

        self.public_key = None
        self.padding_type = SignatureType.PSS
        self.is_debug = False
        self.sat_base_url = PLAYGROUND_SAT_BASE_URL
        self.access_token_base_url = ACCESS_TOKEN_URL
        self.timeout = 30

    def with_logger(self, logger: logging.Logger):
        self.logger = logger
        return self

    def with_public_key(self, public_key: str):
        self.public_key = public_key
        return self

    def with_padding_type(self, padding_type: SignatureType):
        self.padding_type = padding_type
        return self

    def with_is_debug(self, is_debug: bool):
        self.is_debug = is_debug
        return self

    def with_sat_base_url(self, sat_base_url: str):
        self.sat_base_url = sat_base_url
        return self

    def with_timeout(self, timeout: int):
        self.timeout = timeout
        return self

    def with_access_token_base_url(self, access_token_base_url: str):
        self.access_token_base_url = access_token_base_url
        return self


class SATClient:
    """
    SATClient is a main class for SAT SDK, this class contains all the method to interact with SAT service
    """

    _config: SATClientConfig
    signature: Signature
    _logger: logging.Logger
    _http_client: HTTPClient

    def __init__(self, config: SATClientConfig):
        self._config = config
        self.signature = Signature(
            config.private_key, config.public_key, config.padding_type
        )
        self._logger = config.logger
        self._http_client = HTTPClient(
            base_url=config.sat_base_url,
            oauth_base_url=config.access_token_base_url,
            client_id=config.client_id,
            client_secret=config.client_secret,
            logger=config.logger,
        )

    def ping(self) -> Union[PingResponse, ErrorResponse]:
        """
        Ping is a method to check the SAT server health
               :return: PingResponse or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when pinging
        """
        try:
            url = f"{self._config.sat_base_url}{PING_PATH}"

            req = requests.Request(method="GET", url=url)
            response = self._http_client.send_request(req)
            response.raise_for_status()

            json_response = response.json()
            return PingResponse.from_dict(json_response).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when pinging: {exc}")
            raise GeneralException(exc)

    def inquiry(self, req: InquiryRequest) -> Union[InquiryResponse, ErrorResponse]:
        """
        Inquiry is a method to get user bills based on client number and product code
               :param req: InquiryRequest
               :return:  InquiryBaseResponse or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when inquiring
        """
        try:
            url = f"{self._config.sat_base_url}{INQUIRY_PATH}"

            body = generate_json_api_request(req.to_dict())
            http_req = requests.Request(method="POST", url=url, json=body)
            response = self._http_client.send_request(http_req)
            response.raise_for_status()

            json_response = response.json()
            data = parse_json_api_response(json_response)

            return InquiryResponse.from_dict(data).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when inquiry: {exc}")
            raise GeneralException(exc)

    def checkout(self, req: OrderRequest) -> Union[OrderDetail, ErrorResponse]:
        """
        Checkout is a method to do payment an order based on client number, product code and request id.
        Request ID should use unique identifier for each transaction
               :param req: OrderRequest
               :return: OrderDetail or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when checking out
        """
        try:
            url = f"{self._config.sat_base_url}{CHECKOUT_PATH}"

            body = generate_json_api_request(req.to_dict())
            body_str = json.dumps(body)
            signature = self.signature.sign(body_str)

            http_req = requests.Request(
                method="POST",
                url=url,
                json=body,
                headers={"signature": signature},
            )
            response = self._http_client.send_request(http_req)
            response.raise_for_status()

            json_response = response.json()
            data = parse_json_api_response(json_response)

            return OrderDetail.from_dict(data).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when checkout: {exc}")
            raise GeneralException(exc)

    def check_status(self, request_id: str) -> Union[OrderDetail, ErrorResponse]:
        """
        CheckStatus is a method to check the final status of an order.
        request id is must be filled
               :param request_id: request id to check the status
               :return: OrderDetail or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when checking status
        """
        try:
            url = f"{self._config.sat_base_url}{CHECK_STATUS_PATH.format(request_id=request_id)}"

            http_req = requests.Request(
                method="GET",
                url=url,
            )
            response = self._http_client.send_request(http_req)
            response.raise_for_status()

            json_response = response.json()
            data = parse_json_api_response(json_response)

            return OrderDetail.from_dict(data).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when check status: {exc}")
            raise GeneralException(exc)

    def list_product(
        self, code: Optional[str] = None
    ) -> Union[ProductListResponse, ErrorResponse]:
        """
        ListProduct is a method to get all the product list enabled on your credentials.
        you can also specify the product code, to get only one product detail.
        specify product code will be very beneficial to sync product status on your engine
        it will come with low bandwidth and fast response
               :param code: product code to filter the product list
               :return: ProductListResponse or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when listing product
        """
        try:
            url = f"{self._config.sat_base_url}{PRODUCT_LIST_PATH}"

            http_req = requests.Request(
                method="GET",
                url=url,
                params={"product_code": code},
            )
            response = self._http_client.send_request(http_req)
            response.raise_for_status()

            json_response = response.json()
            data = parse_json_api_list_response(json_response)
            products = [
                PartnerProduct.from_dict(item).with_raw_response(response)
                for item in data
            ]

            return ProductListResponse(products=products).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when list product: {exc}")
            raise GeneralException(exc)

    def account(self) -> Union[Account, ErrorResponse]:
        """
        Account is a method to check account balance
               :return: Account or ErrorResponse
               :raise ResponseGeneralException: if there are unexpected error when hitting SAT (403 forbidden, network error)
               :raise GeneralException: if there is an unexpected exception when checking account
        """

        try:
            url = f"{self._config.sat_base_url}{ACCOUNT_PATH}"

            http_req = requests.Request(
                method="GET",
                url=url,
            )
            response = self._http_client.send_request(http_req)
            response.raise_for_status()

            json_response = response.json()
            data = parse_json_api_response(json_response)

            return Account.from_dict(data).with_raw_response(response)
        except HTTPError as exc:
            return self._handle_http_error(exc)
        except Exception as exc:
            self._logger.error(f"Error when checking balance: {exc}")
            raise GeneralException(exc)

    def handle_callback(
        self,
        sat_response_data: Dict[str, Any],
        sat_response_headers: Dict[str, Any],
        do: Callable[[OrderDetail], None],
    ):
        """
        HandleCallback is method http.HandlerFunc to handle callback request from SAT
        you can customize the implementation based on this interface Callback
           :param sat_response_data: JSON API response data from SAT webhook, must be a dictionary
           :param sat_response_headers: HTTP headers from SAT webhook, must be a dictionary
           :param do: Your custom function to handle the callback, must be a function with OrderDetail as
                      parameter and return void
           :raise InvalidInputException: if the header does not contain the signature
           :raise UnauthenticatedException: if the signature is not valid
        """
        signature: Optional[str] = sat_response_headers.get(
            "signature"
        ) or sat_response_headers.get("Signature")
        if signature is None:
            raise InvalidInputException(
                "Signature is not present in the header, please check the request"
            )

        verify = self.signature.verify(json.dumps(sat_response_data), signature)
        if not verify:
            raise UnauthenticatedException("Signature is not valid")

        data = parse_json_api_response(sat_response_data)
        order_detail = OrderDetail.from_dict(data)

        do(order_detail)

    def get_http_client(self) -> HTTPClient:
        """
        GetHTTTPClient will return http client which already wrapped to support oauth2
        for custom integration to SAT Service
                :return: HTTPClient
        """
        return self._http_client

    def get_signature(self) -> Signature:
        """
        GetSignature will return signature object which already wrapped to support RSA signature
        for custom integration to SAT Service
                :return: Signature
        """
        return self.signature

    def _handle_http_error(self, exc: HTTPError):
        try:
            status = exc.response.status_code
            message = exc.response.text
            self._logger.debug(f"HTTP Error: {status}, {message}")

            data = exc.response.json()
            resp = ErrorResponse.from_dict(data).with_raw_response(exc.response)
            return resp
        except Exception as e:
            raise ResponseGeneralException(exc)



from typing import Any, Dict, List


def parse_json_api_response(response: dict) -> Dict[str, Any]:
    """
    Parse JSON API response to dictionary

    :param response: JSON API response
    :return: aggregate dictionary of id, type, and attributes
    """
    data = response.get("data", {})
    return __parse_json_api_dict(data)


def parse_json_api_list_response(response: dict) -> List[Dict[str, Any]]:
    """
    Parse JSON API list response to list of dictionary

    :param response: JSON API response
    :return: aggregate dictionary of id, type, and attributes
    """
    data = response.get("data", [])

    return [__parse_json_api_dict(item) for item in data]


def __parse_json_api_dict(data):
    id = data.get("id", "")
    type = data.get("type", "")
    attributes = data.get("attributes", {})
    return {"id": id, "type": type, **attributes}


def generate_json_api_request(request: dict) -> Dict[str, Any]:
    """
    Generate JSON API request from dictionary

    :param request: normal dictionary
    :return: JSON API request format
    """
    id = extract_id(request)
    type = extract_type(request)

    data = {}
    if id is not None:
        data["id"] = id
    if type is not None:
        data["type"] = type

    # Omit none values
    attribute = {k: v for k, v in request.items() if v is not None}
    data["attributes"] = attribute

    return {"data": data}


def extract_type(request):
    type = None
    if "type" in request:
        type = request.get("type", None)
        del request["type"]
    return type


def extract_id(request):
    id = None
    if "id" in request:
        id = request.get("id", None)
        del request["id"]
    if "request_id" in request:
        id = request.get("request_id", None)
        del request["request_id"]
    return id



import logging
from datetime import datetime, timezone

import requests
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from py_sat.constant import DATE_TIME_FORMAT, SDK_LABEL


class HTTPClient:
    _base_url: str
    _oauth_base_url: str
    _access_token: str
    _client_id: str
    _client_secret: str
    _session: requests.Session
    _auth: OAuth2Session
    _logger: logging.Logger
    _is_debug: bool

    def __init__(
        self,
        base_url: str,
        oauth_base_url: str,
        client_id: str,
        client_secret: str,
        logger: logging.Logger,
        is_debug: bool = False,
    ):
        self._base_url = base_url
        self._oauth_base_url = oauth_base_url
        self._client_id = client_id
        self._client_secret = client_secret
        self._session = self._prepare_session()
        self._logger = logger
        self._is_debug = is_debug

        self._auth = OAuth2Session(
            client=BackendApplicationClient(client_id),
            client_id=client_id,
            token_updater=lambda token: self._save_token(token),
            auto_refresh_url=self._oauth_base_url,
            auto_refresh_kwargs={
                "client_id": client_id,
                "client_secret": client_secret,
            },
        )
        self._access_token = self.get_access_token()

    def _save_token(self, token):
        self._access_token = token

    @staticmethod
    def _prepare_session():
        session = requests.Session()
        session.headers.update({})

        return session

    def get_access_token(self):
        token = self._auth.fetch_token(
            token_url=self._oauth_base_url,
            client_id=self._client_id,
            client_secret=self._client_secret,
        )

        return token.get("access_token")

    def send_request(self, request: requests.Request) -> requests.Response:
        """Sends a prepared Request object and returns the response."""
        request.headers["Date"] = datetime.now(timezone.utc).strftime(DATE_TIME_FORMAT)
        request.headers["Content-Type"] = "application/json"
        request.headers["Accept"] = "application/json"
        request.headers["X-Sat-Sdk-Version"] = SDK_LABEL
        request.headers["Authorization"] = f"Bearer {self._access_token}"

        prepared_request = self._auth.prepare_request(request)

        response = self._session.send(prepared_request)

        self._logger.info(
            f"Request: {response.request.url}, {response.request.method} {response.request.body} {response.request.headers}"
        )

        return response



from abc import ABC, abstractmethod
from typing import Union

from Crypto.PublicKey import RSA


class SignatureAlgorithm(ABC):
    @abstractmethod
    def verify(
        self, public_key: RSA.RsaKey, msg: Union[str, bytes], signature: str
    ) -> bool:
        """Verifies a signature"""
        pass

    @abstractmethod
    def sign(self, private_key: RSA.RsaKey, msg: Union[str, bytes]) -> str:
        """Signs a message"""
        pass



from py_sat.models.account import Account
from py_sat.models.base import Field
from py_sat.models.error import ErrorObject, ErrorResponse
from py_sat.models.inquiry import InquiryRequest, InquiryResponse
from py_sat.models.order import OrderDetail, OrderRequest
from py_sat.models.ping import PingResponse
from py_sat.models.product import (PartnerProduct, ProductListResponse,
                                   ProductStatus)



from dataclasses import dataclass, field
from typing import List

from dataclasses_json import Undefined, dataclass_json

from py_sat.models.base import BaseResponse


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class ErrorObject:
    """
    Error object, represent error detail from SAT API
    """

    id: str = field(default="")
    title: str = field(default="")
    detail: str = field(default="")
    status: str = field(default="")
    code: str = field(default="")
    meta: dict = field(default_factory=dict)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class ErrorResponse(BaseResponse):
    """
    Error response object, represent error response from SAT API
    """

    errors: List[ErrorObject] = field(default_factory=list)

    def get_error_messages(self) -> str:
        """
        Get error messages from errors
        :return: string of error messages, separated by new line
        """
        if len(self.errors) <= 0:
            return ""

        messages = [
            f"{error.status} - {error.code} - {error.detail}" for error in self.errors
        ]
        return "\n".join(messages)

    def get_error_codes(self) -> str:
        """
        Get error codes from errors
        :return: string of error codes, separated by comma
        """
        if len(self.errors) <= 0:
            return ""

        return ", ".join([error.code for error in self.errors])

    def get_error_statuses(self) -> str:
        """
        Get error statuses from errors
        :return: string of error statuses, separated by comma
        """
        if len(self.errors) <= 0:
            return ""

        return ", ".join([error.status for error in self.errors])

    def get_error_details(self) -> str:
        """
        Get error details from errors
        :return: string of error details, separated by comma
        """

        if len(self.errors) <= 0:
            return ""

        return ", ".join([error.detail for error in self.errors])



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from dataclasses_json import Undefined, config, dataclass_json

from py_sat.models.base import (BaseRequest, BaseResponse, Field,
                                datetime_config)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class OrderRequest(BaseRequest):
    id: str = field(default="")
    product_code: str = field(default="")
    client_number: str = field(default="")
    amount: float = field(default=0.0)
    fields: Optional[List[Field]] = field(default=None)
    downline_id: Optional[str] = field(default=None)

    type: str = field(default="order", init=False)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class OrderDetail(BaseResponse):
    id: str = field(default="")
    fields: Optional[List[Field]] = field(default=None)
    fulfillment_result: List[Field] = field(default_factory=list)
    fulfilled_at: Optional[datetime] = field(
        default=None,
        metadata=datetime_config,
    )
    error_code: str = field(default="")
    error_detail: str = field(default="")
    product_code: str = field(default="")
    status: str = field(default="")
    partner_fee: int = field(default=0)
    sales_price: int = field(default=0)
    admin_fee: int = field(default=0)
    client_name: str = field(default="")
    client_number: str = field(default="")
    voucher_code: str = field(default="")
    serial_number: str = field(default="")



from dataclasses import dataclass, field

from dataclasses_json import Undefined, dataclass_json

from py_sat.models.base import BaseResponse


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Account(BaseResponse):
    """
    Account is a schema related with account entity
    """

    id: int
    saldo: int



from enum import Enum
from typing import Optional

from Crypto.PublicKey import RSA

from py_sat.exceptions import InvalidInputException, SignatureErrorException
from py_sat.signature.interface import SignatureAlgorithm
from py_sat.signature.pss import PSSPaddingAlgorithm


class SignatureType(Enum):
    PSS = "PSS"


class Signature:
    """
    Signature to hold that signature needs, and contain parsed public and private key
    """

    _private_key: Optional[RSA.RsaKey]
    _public_key: Optional[RSA.RsaKey]
    _algorithm: SignatureAlgorithm

    def __init__(
        self,
        private_key_str: Optional[str],
        public_key_str: Optional[str],
        padding_type: SignatureType,
    ):
        if not padding_type:
            raise InvalidInputException("Padding type is required")

        self._private_key = self._parse_rsa_private_key_from_pem_str(private_key_str)
        self._public_key = self._parse_public_key(public_key_str)
        self._algorithm = self.__decide_padding_algorithm(padding_type)

    def verify(self, msg: str, signature: str) -> bool:
        """
        Verify the signature of the message, return True if the signature is valid, False otherwise

        :param msg: message to verify, must be a string
        :param signature: signature to verify, must be a string
        :return: True if the signature is valid, False otherwise
        :raise InvalidInputException: if the message or signature is not a string or public key is not set
        :raise SignatureErrorException: if there is an error verifying the signature
        """
        if not isinstance(msg, str) or not isinstance(signature, str):
            raise InvalidInputException("Message and signature must be strings")

        if not self._public_key:
            raise InvalidInputException("Public key not set")

        try:
            return self._algorithm.verify(self._public_key, msg, signature)
        except Exception as exc:
            raise SignatureErrorException(f"Error verifying signature: {exc}")

    def sign(self, msg: str) -> str:
        """
        Sign the message and return the signature

        :param msg: to sign, must be a string
        :return: the signature as a string
        :raises InvalidInputException: if the message is not a string or private key is not set
        :raises SignatureErrorException: if there is an error signing the message
        """
        if not isinstance(msg, str):
            raise InvalidInputException("Message must be a string")

        if not self._private_key:
            raise InvalidInputException("Private key not set")

        try:
            return self._algorithm.sign(self._private_key, msg)
        except Exception as exc:
            raise SignatureErrorException(f"Error signing message: {exc}")

    @staticmethod
    def __decide_padding_algorithm(padding_type: SignatureType) -> SignatureAlgorithm:
        if padding_type == SignatureType.PSS:
            return PSSPaddingAlgorithm()
        else:
            raise InvalidInputException(f"Unknown padding type: {padding_type}")

    @staticmethod
    def _parse_rsa_private_key_from_pem_str(
        private_key_pem: Optional[str],
    ) -> Optional[RSA.RsaKey]:
        try:
            """Parses an RSA private key from a PEM-encoded string."""
            if not private_key_pem:
                return None

            private_key = RSA.import_key(private_key_pem.encode())
            if not isinstance(private_key, RSA.RsaKey):
                raise InvalidInputException("Key is not a valid RSA public key")

            return private_key
        except (ValueError, IndexError, TypeError) as exc:
            raise InvalidInputException(f"Invalid RSA private key PEM: {exc}")

    @staticmethod
    def _parse_public_key(public_key_pem: Optional[str]) -> Optional[RSA.RsaKey]:
        """Parses an RSA public key from a PEM-encoded string."""
        try:
            if not public_key_pem:
                return None

            public_key = RSA.import_key(public_key_pem.encode())

            if not isinstance(public_key, RSA.RsaKey):
                raise InvalidInputException("Key is not a valid RSA public key")
            return public_key
        except (ValueError, IndexError, TypeError) as exc:
            raise InvalidInputException(f"Invalid RSA public key PEM: {exc}")



import enum
from dataclasses import dataclass, field
from typing import List

from dataclasses_json import Undefined, dataclass_json

from py_sat.models.base import BaseResponse


class ProductStatus(enum.Enum):
    Active = 1
    Inactive = 2


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class PartnerProduct(BaseResponse):
    id: str = field(default="")
    name: str = field(default="")
    operator_name: str = field(default="")
    category_name: str = field(default="")
    is_inquiry: bool = field(default=False)
    sales_price: int = field(default=0)
    status: ProductStatus = field(default=ProductStatus.Active)
    client_number: str = field(default="")


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class ProductListResponse(BaseResponse):
    products: List[PartnerProduct] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import Undefined, dataclass_json

from py_sat.models.base import BaseRequest, BaseResponse, Field


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class InquiryRequest(BaseRequest):
    """
    InquiryRequest to hold inquiry request data
    """

    product_code: str
    client_number: str
    amount: Optional[int] = field(default=None)
    id: Optional[str] = field(default=None)
    fields: Optional[List[Field]] = field(default=None)
    downline_id: Optional[str] = field(default=None)
    type: str = field(default="inquiry", init=False)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class InquiryResponse(BaseResponse):
    """
    InquiryResponse to hold inquiry response
    """

    id: str = field(default="")
    product_code: str = field(default="")
    sales_price: float = field(default=0.0)
    fields: List[Field] = field(default_factory=list)
    inquiry_result: List[Field] = field(default_factory=list)
    base_price: float = field(default=0.0)
    admin_fee: float = field(default=0.0)
    client_name: str = field(default="")
    client_number: str = field(default="")
    meter_id: str = field(default="")
    ref_id: str = field(default="")
    max_payment: int = field(default=0)
    min_payment: int = field(default=0)
